@@ -11,7 +11,7 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * RCS: @(#) $Id: tclEvent.c,v 1.28.2.8 2004/07/21 01:30:57 hobbs Exp $
+ * RCS: @(#) $Id: tclEvent.c,v 1.28.2.13 2005/08/16 15:23:56 kennykb Exp $
  */
 
 #include "tclInt.h"
@@ -783,6 +783,29 @@ void
 Tcl_Finalize()
 {
     ExitHandler *exitPtr;
+    
+    /*
+     * Invoke exit handlers first.
+     */
+
+    Tcl_MutexLock(&exitMutex);
+    inFinalize = 1;
+    for (exitPtr = firstExitPtr; exitPtr != NULL; exitPtr = firstExitPtr) {
+	/*
+	 * Be careful to remove the handler from the list before
+	 * invoking its callback.  This protects us against
+	 * double-freeing if the callback should call
+	 * Tcl_DeleteExitHandler on itself.
+	 */
+
+	firstExitPtr = exitPtr->nextPtr;
+	Tcl_MutexUnlock(&exitMutex);
+	(*exitPtr->proc)(exitPtr->clientData);
+	ckfree((char *) exitPtr);
+	Tcl_MutexLock(&exitMutex);
+    }    
+    firstExitPtr = NULL;
+    Tcl_MutexUnlock(&exitMutex);
 
     TclpInitLock();
     if (subsystemsInitialized != 0) {
@@ -795,29 +818,6 @@ Tcl_Finalize()
 
 	(void) TCL_TSD_INIT(&dataKey);
 
-	/*
-	 * Invoke exit handlers first.
-	 */
-
-	Tcl_MutexLock(&exitMutex);
-	inFinalize = 1;
-	for (exitPtr = firstExitPtr; exitPtr != NULL; exitPtr = firstExitPtr) {
-	    /*
-	     * Be careful to remove the handler from the list before
-	     * invoking its callback.  This protects us against
-	     * double-freeing if the callback should call
-	     * Tcl_DeleteExitHandler on itself.
-	     */
-
-	    firstExitPtr = exitPtr->nextPtr;
-	    Tcl_MutexUnlock(&exitMutex);
-	    (*exitPtr->proc)(exitPtr->clientData);
-	    ckfree((char *) exitPtr);
-	    Tcl_MutexLock(&exitMutex);
-	}    
-	firstExitPtr = NULL;
-	Tcl_MutexUnlock(&exitMutex);
-
 	/*
 	 * Clean up after the current thread now, after exit handlers.
 	 * In particular, the testexithandler command sets up something
@@ -833,15 +833,29 @@ Tcl_Finalize()
 	 * order dependencies.
 	 */
 
-	TclFinalizeCompExecEnv();
+	TclFinalizeCompilation();
+	TclFinalizeExecution();
 	TclFinalizeEnvironment();
 
 	/* 
 	 * Finalizing the filesystem must come after anything which
 	 * might conceivably interact with the 'Tcl_FS' API. 
 	 */
+
 	TclFinalizeFilesystem();
 
+	/*
+	 * Undo all the Tcl_ObjType registrations, and reset the master list
+	 * of free Tcl_Obj's.  After this returns, no more Tcl_Obj's should
+	 * be allocated or freed.
+	 *
+	 * Note in particular that TclFinalizeObjects() must follow
+	 * TclFinalizeFilesystem() because TclFinalizeFilesystem free's
+	 * the Tcl_Obj that holds the path of the current working directory.
+	 */
+
+	TclFinalizeObjects();
+
 	/* 
 	 * We must be sure the encoding finalization doesn't need
 	 * to examine the filesystem in any way.  Since it only
@@ -869,6 +883,24 @@ Tcl_Finalize()
 	
 	Tcl_SetPanicProc(NULL);
 
+	/*
+	 * There have been several bugs in the past that cause
+	 * exit handlers to be established during Tcl_Finalize
+	 * processing.  Such exit handlers leave malloc'ed memory,
+	 * and Tcl_FinalizeThreadAlloc or Tcl_FinalizeMemorySubsystem
+	 * will result in a corrupted heap.  The result can be a
+	 * mysterious crash on process exit.  Check here that
+	 * nobody's done this.
+	 */
+
+#ifdef TCL_MEM_DEBUG
+	if ( firstExitPtr != NULL ) {
+	    Tcl_Panic( "exit handlers were created during Tcl_Finalize" );
+	}
+#endif
+
+	TclFinalizePreserve();
+
 	/*
 	 * Free synchronization objects.  There really should only be one
 	 * thread alive at this moment.
@@ -876,6 +908,10 @@ Tcl_Finalize()
 
 	TclFinalizeSynchronization();
 
+#if defined(TCL_THREADS) && defined(USE_THREAD_ALLOC) && !defined(TCL_MEM_DEBUG) && !defined(PURIFY)
+	TclFinalizeThreadAlloc();
+#endif
+
 	/*
 	 * We defer unloading of packages until very late 
 	 * to avoid memory access issues.  Both exit callbacks and
@@ -895,11 +931,9 @@ Tcl_Finalize()
 	TclResetFilesystem();
 	
 	/*
-	 * There shouldn't be any malloc'ed memory after this.
+	 * At this point, there should no longer be any ckalloc'ed memory.
 	 */
-#if defined(TCL_THREADS) && defined(USE_THREAD_ALLOC) && !defined(TCL_MEM_DEBUG) && !defined(PURIFY)
-	TclFinalizeThreadAlloc();
-#endif
+
 	TclFinalizeMemorySubsystem();
 	inFinalize = 0;
     }
