@@ -11,7 +11,7 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * RCS: @(#) $Id: tclWinFile.c,v 1.44.2.10 2004/10/08 20:16:42 hobbs Exp $
+ * RCS: @(#) $Id: tclWinFile.c,v 1.44.2.13 2005/11/15 16:41:41 kennykb Exp $
  */
 
 //#define _WIN32_WINNT  0x0500
@@ -22,6 +22,13 @@
 #include <shlobj.h>
 #include <lmaccess.h>		/* For TclpGetUserHome(). */
 
+/*
+ * The number of 100-ns intervals between the Windows system epoch (1601-01-01
+ * on the proleptic Gregorian calendar) and the Posix epoch (1970-01-01).
+ */
+
+#define POSIX_EPOCH_AS_FILETIME		116444736000000000
+
 /*
  * Declarations for 'link' related information.  This information
  * should come with VC++ 6.0, but is not in some older SDKs.
@@ -151,6 +158,7 @@ typedef enum _FINDEX_SEARCH_OPS {
 /* Other typedefs required by this code */
 
 static time_t		ToCTime(FILETIME fileTime);
+static void		FromCTime(time_t posixTime, FILETIME *fileTime);
 
 typedef NET_API_STATUS NET_API_FUNCTION NETUSERGETINFOPROC
 	(LPWSTR servername, LPWSTR username, DWORD level, LPBYTE *bufptr);
@@ -178,6 +186,7 @@ static int NativeWriteReparse(CONST TCHAR* LinkDirectory,
 static int NativeMatchType(int isDrive, DWORD attr, CONST TCHAR* nativeName, 
 			   Tcl_GlobTypeData *types);
 static int WinIsDrive(CONST char *name, int nameLen);
+static int WinIsReserved(CONST char *path);
 static Tcl_Obj* WinReadLink(CONST TCHAR* LinkSource);
 static Tcl_Obj* WinReadLinkDirectory(CONST TCHAR* LinkDirectory);
 static int WinLink(CONST TCHAR* LinkSource, CONST TCHAR* LinkTarget, 
@@ -1010,6 +1019,52 @@ WinIsDrive(
     return 0;
 }
 
+/* 
+ * Does the given path represent a reserved window path name?  If not
+ * return 0, if true, return the number of characters of the path that
+ * we actually want (not any trailing :).
+ */
+static int WinIsReserved(
+   CONST char *path)    /* Path in UTF-8  */
+{
+    if ((path[0] == 'c' || path[0] == 'C') 
+	&& (path[1] == 'o' || path[1] == 'O')) {
+	if ((path[2] == 'm' || path[2] == 'M')
+	    && path[3] >= '1' && path[3] <= '4') {
+	    /* May have match for 'com[1-4]:?', which is a serial port */
+	    if (path[4] == '\0') {
+		return 4;
+	    } else if (path [4] == ':' && path[5] == '\0') {
+		return 4;
+	    }
+	} else if ((path[2] == 'n' || path[2] == 'N') && path[3] == '\0') {
+	    /* Have match for 'con' */
+	    return 3;
+	}
+    } else if ((path[0] == 'l' || path[0] == 'L')
+	       && (path[1] == 'p' || path[1] == 'P')
+	       && (path[2] == 't' || path[2] == 'T')) {
+	if (path[3] >= '1' && path[3] <= '3') {
+	    /* May have match for 'lpt[1-3]:?' */
+	    if (path[4] == '\0') {
+		return 4;
+	    } else if (path [4] == ':' && path[5] == '\0') {
+		return 4;
+	    }
+	}
+    } else if (stricmp(path, "prn") == 0) {
+	/* Have match for 'prn' */
+	return 3;
+    } else if (stricmp(path, "nul") == 0) {
+	/* Have match for 'nul' */
+	return 3;
+    } else if (stricmp(path, "aux") == 0) {
+	/* Have match for 'aux' */
+	return 3;
+    }
+    return 0;
+}
+
 /*
  *----------------------------------------------------------------------
  * 
@@ -1813,57 +1868,57 @@ NativeStatMode(DWORD attr, int checkLinks, int isExec)
     return (unsigned short)mode;
 }
 
+/*
+ *------------------------------------------------------------------------
+ *
+ * ToCTime --
+ *
+ *	Converts a Windows FILETIME to a time_t in UTC.
+ *
+ * Results:
+ *	Returns the count of seconds from the Posix epoch.
+ *
+ *------------------------------------------------------------------------
+ */
+
 static time_t
 ToCTime(
-    FILETIME fileTime)		/* UTC Time to convert to local time_t. */
+    FILETIME fileTime)		/* UTC time */
 {
-    FILETIME localFileTime;
-    SYSTEMTIME systemTime;
-    struct tm tm;
+    LARGE_INTEGER convertedTime;
 
-    if (FileTimeToLocalFileTime(&fileTime, &localFileTime) == 0) {
-	return 0;
-    }
-    if (FileTimeToSystemTime(&localFileTime, &systemTime) == 0) {
-	return 0;
-    }
-    tm.tm_sec = systemTime.wSecond;
-    tm.tm_min = systemTime.wMinute;
-    tm.tm_hour = systemTime.wHour;
-    tm.tm_mday = systemTime.wDay;
-    tm.tm_mon = systemTime.wMonth - 1;
-    tm.tm_year = systemTime.wYear - 1900;
-    tm.tm_wday = 0;
-    tm.tm_yday = 0;
-    tm.tm_isdst = -1;
+    convertedTime.LowPart = fileTime.dwLowDateTime;
+    convertedTime.HighPart = (LONG) fileTime.dwHighDateTime;
 
-    return mktime(&tm);
+    return (time_t) ((convertedTime.QuadPart
+	    - (Tcl_WideInt) POSIX_EPOCH_AS_FILETIME) / (Tcl_WideInt) 10000000);
 }
+
+/*
+ *------------------------------------------------------------------------
+ *
+ * FromCTime --
+ *
+ *	Converts a time_t to a Windows FILETIME
+ *
+ * Results:
+ *	Returns the count of 100-ns ticks seconds from the Windows epoch.
+ *
+ *------------------------------------------------------------------------
+ */
 
-#if 0
-
-    /*
-     * Borland's stat doesn't take into account localtime.
-     */
-
-    if ((result == 0) && (buf->st_mtime != 0)) {
-	TIME_ZONE_INFORMATION tz;
-	int time, bias;
-
-	time = GetTimeZoneInformation(&tz);
-	bias = tz.Bias;
-	if (time == TIME_ZONE_ID_DAYLIGHT) {
-	    bias += tz.DaylightBias;
-	}
-	bias *= 60;
-	buf->st_atime -= bias;
-	buf->st_ctime -= bias;
-	buf->st_mtime -= bias;
-    }
-
-#endif
-
-
+static void
+FromCTime(
+    time_t posixTime,
+    FILETIME* fileTime)		/* UTC Time */
+{
+    LARGE_INTEGER convertedTime;
+    convertedTime.QuadPart = ((LONGLONG) posixTime) * 10000000
+	+ POSIX_EPOCH_AS_FILETIME;
+    fileTime->dwLowDateTime = convertedTime.LowPart;
+    fileTime->dwHighDateTime = convertedTime.HighPart;
+}
+
 #if 0
 /*
  *-------------------------------------------------------------------------
@@ -2038,7 +2093,7 @@ TclpFilesystemPathType(pathObjPtr)
 {
 #define VOL_BUF_SIZE 32
     int found;
-    char volType[VOL_BUF_SIZE];
+    WCHAR volType[VOL_BUF_SIZE];
     char* firstSeparator;
     CONST char *path;
     
@@ -2067,7 +2122,7 @@ TclpFilesystemPathType(pathObjPtr)
 	Tcl_DString ds;
 	Tcl_Obj *objPtr;
 	
-	Tcl_WinTCharToUtf(volType, -1, &ds);
+	Tcl_WinTCharToUtf((CONST char *)volType, -1, &ds);
 	objPtr = Tcl_NewStringObj(Tcl_DStringValue(&ds),Tcl_DStringLength(&ds));
 	Tcl_DStringFree(&ds);
 	return objPtr;
@@ -2140,9 +2195,22 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
 		 * the current normalized path, if the file exists.
 		 */
 		if (isDrive) {
-		    if (GetFileAttributesA(nativePath) 
-			== 0xffffffff) {
+		    if (GetFileAttributesA(nativePath) == 0xffffffff) {
 			/* File doesn't exist */
+			if (isDrive) {
+			    int len = WinIsReserved(path);
+			    if (len > 0) {
+				/* Actually it does exist - COM1, etc */
+				int i;
+				for (i=0;i<len;i++) {
+				    if (nativePath[i] >= 'a') {
+					((char*)nativePath)[i] -= ('a' - 'A');
+				    }
+				}
+				Tcl_DStringAppend(&dsNorm, nativePath, len);
+				lastValidPathEnd = currentPathEndPosition;
+			    }
+			}
 			Tcl_DStringFree(&ds);
 			break;
 		    }
@@ -2209,6 +2277,23 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
 		if ((*tclWinProcs->getFileAttributesExProc)(nativePath,
 		    GetFileExInfoStandard, &data) != TRUE) {
 		    /* File doesn't exist */
+		    if (isDrive) {
+			int len = WinIsReserved(path);
+			if (len > 0) {
+			    /* Actually it does exist - COM1, etc */
+			    int i;
+			    for (i=0;i<len;i++) {
+				WCHAR wc = ((WCHAR*)nativePath)[i];
+				if (wc >= L'a') {
+				    wc -= (L'a' - L'A');
+				    ((WCHAR*)nativePath)[i] = wc;
+				}
+			    }
+			    Tcl_DStringAppend(&dsNorm, nativePath,
+					      sizeof(WCHAR)*len);
+			    lastValidPathEnd = currentPathEndPosition;
+			}
+		    }
 		    Tcl_DStringFree(&ds);
 		    break;
 		}
@@ -2366,6 +2451,7 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
     return nextCheckpoint;
 }
 
+
 /*
  *---------------------------------------------------------------------------
  *
@@ -2377,25 +2463,41 @@ TclpObjNormalizePath(interp, pathPtr, nextCheckpoint)
  *	0 on success, -1 on error.
  *
  * Side effects:
- *	None.
+ *	Sets errno to a representation of any Windows problem that's observed
+ *	in the process.
  *
  *---------------------------------------------------------------------------
  */
+
 int
-TclpUtime(pathPtr, tval)
-    Tcl_Obj *pathPtr;      /* File to modify */
-    struct utimbuf *tval;  /* New modification date structure */
+TclpUtime(
+    Tcl_Obj *pathPtr,		/* File to modify */
+    struct utimbuf *tval)	/* New modification date structure */
 {
-    int res;
-    /* 
-     * Windows uses a slightly different structure name and, possibly,
-     * contents, so we have to copy the information over
+    int res = 0;
+    HANDLE fileHandle;
+    FILETIME lastAccessTime, lastModTime;
+
+    FromCTime(tval->actime, &lastAccessTime);
+    FromCTime(tval->modtime, &lastModTime);
+
+    /*
+     * We use the native APIs (not 'utime') because there are some daylight
+     * savings complications that utime gets wrong.
      */
-    struct _utimbuf buf;
-    
-    buf.actime = tval->actime;
-    buf.modtime = tval->modtime;
-    
-    res = (*tclWinProcs->utimeProc)(Tcl_FSGetNativePath(pathPtr),&buf);
+
+    fileHandle = (tclWinProcs->createFileProc) (
+	    (CONST TCHAR *) Tcl_FSGetNativePath(pathPtr),
+	    FILE_WRITE_ATTRIBUTES, 0, NULL, OPEN_EXISTING,
+	    FILE_ATTRIBUTE_NORMAL, NULL);
+
+    if (fileHandle == INVALID_HANDLE_VALUE ||
+	    !SetFileTime(fileHandle, NULL, &lastAccessTime, &lastModTime)) {
+	TclWinConvertError(GetLastError());
+	res = -1;
+    }
+    if (fileHandle != INVALID_HANDLE_VALUE) {
+	CloseHandle(fileHandle);
+    }
     return res;
 }
